package ru.dbotthepony.mc.otm.client.screen.panels.button

import com.mojang.blaze3d.platform.InputConstants
import com.mojang.blaze3d.vertex.PoseStack
import net.minecraft.ChatFormatting
import net.minecraft.client.gui.screens.Screen
import net.minecraft.network.chat.Component
import ru.dbotthepony.mc.otm.client.render.AbstractMatterySprite
import ru.dbotthepony.mc.otm.client.render.UVWindingOrder
import ru.dbotthepony.mc.otm.client.screen.panels.EditablePanel
import ru.dbotthepony.mc.otm.core.GetterSetter
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.core.next
import ru.dbotthepony.mc.otm.core.prev
import ru.dbotthepony.mc.otm.core.util.EnumValueCodec
import ru.dbotthepony.mc.otm.core.value
import java.util.*
import kotlin.collections.ArrayList

abstract class EnumRectangleButtonPanel<out S : Screen, T : Enum<T>>(
	screen: S,
	parent: EditablePanel<*>?,
	x: Float = 0f,
	y: Float = 0f,
	width: Float,
	height: Float,
	enum: Class<T>,
	val prop: GetterSetter<T>,
	val defaultValue: T,
) : RectangleButtonPanel<S>(screen, parent, x, y, width, height, null) {
	val enum = EnumValueCodec.searchClass(enum)
	private var isBuilding = true

	data class Entry<T : Enum<T>>(
		val key: T,
		val skinElement: AbstractMatterySprite?,
		val winding: UVWindingOrder = UVWindingOrder.NORMAL,
		val tooltip: Component? = null
	)

	protected val enumMapping = EnumMap<T, Entry<T>>(enum)

	fun isFullyDefined(): Boolean {
		if (!isBuilding) return true

		for (value in enum.enumConstants) {
			if (enumMapping[value] == null) {
				return false
			}
		}

		return true
	}

	val missingValues: List<T> get() {
		val missing = ArrayList<T>()

		for (value in enum.enumConstants) {
			if (enumMapping[value] == null) {
				missing.add(value)
			}
		}

		return missing
	}

	fun add(key: T, skinElement: AbstractMatterySprite? = null, tooltip: Component? = null, winding: UVWindingOrder = UVWindingOrder.NORMAL): EnumRectangleButtonPanel<S, T> {
		return add(Entry(key = key, skinElement = skinElement, winding = winding, tooltip = tooltip))
	}

	fun add(entry: Entry<T>): EnumRectangleButtonPanel<S, T> {
		check(isBuilding) { "Not building" }
		check(enumMapping.put(entry.key, entry) == null) { "Already has mapping for ${entry.key}" }
		if (enumMapping.size == enum.enumConstants.size) finish()
		return this
	}

	fun finish(): EnumRectangleButtonPanel<S, T> {
		check(isBuilding) { "Not building" }
		check(isFullyDefined()) { "Not all enums having their mapping defined, missing are: ${missingValues.joinToString(", ")}" }
		isBuilding = false
		return this
	}

	override fun innerRender(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float) {
		check(!isBuilding) { "Still building this button!" }
		super.innerRender(stack, mouseX, mouseY, partialTick)
		val entry = checkNotNull(enumMapping[prop.get()]) { "HOW" }
		entry.skinElement?.render(stack, 0f, 0f, width, height, entry.winding)
	}

	override fun mouseClickedInner(x: Double, y: Double, button: Int): Boolean {
		check(!isBuilding) { "Still building this button!" }
		return super.mouseClickedInner(x, y, button)
	}

	override fun mouseReleasedInner(x: Double, y: Double, button: Int): Boolean {
		check(!isBuilding) { "Still building this button!" }
		return super.mouseReleasedInner(x, y, button)
	}

	override fun test(value: Int): Boolean {
		return value == InputConstants.MOUSE_BUTTON_LEFT ||
				value == InputConstants.MOUSE_BUTTON_RIGHT ||
				value == InputConstants.MOUSE_BUTTON_MIDDLE
	}

	override fun onClick(mouseButton: Int) {
		when (mouseButton) {
			InputConstants.MOUSE_BUTTON_LEFT -> prop.value = prop.value.next(enum.enumConstants)
			InputConstants.MOUSE_BUTTON_RIGHT -> prop.value = prop.value.prev(enum.enumConstants)
			InputConstants.MOUSE_BUTTON_MIDDLE -> {
				if (prop.value != defaultValue) {
					prop.value = defaultValue
				}
			}
		}
	}

	override fun innerRenderTooltips(stack: PoseStack, mouseX: Float, mouseY: Float, partialTick: Float): Boolean {
		if (!isHovered && !isGrabbingMouseInput()) {
			return super.innerRenderTooltips(stack, mouseX, mouseY, partialTick)
		}

		if (tooltip == null && tooltipList == null && enumMapping.values.none { it.tooltip != null }) {
			return super.innerRenderTooltips(stack, mouseX, mouseY, partialTick)
		}

		val listing = ArrayList<Component>()

		if (tooltipList != null) {
			listing.addAll(tooltipList!!)
			listing.add(SPACE)
		} else if (tooltip != null) {
			listing.add(tooltip!!)
			listing.add(SPACE)
		}

		for (entry in enumMapping.values) {
			if (entry.tooltip != null) {
				listing.add(entry.tooltip.copy().withStyle(if (entry.key == prop.get()) ChatFormatting.WHITE else ChatFormatting.GRAY))
			}
		}

		screen.renderComponentTooltip(stack, listing, mouseX.toInt(), mouseY.toInt(), font)
		return true
	}

	companion object {
		private val SPACE = TextComponent("")
	}
}
