package ru.dbotthepony.mc.otm.core.util

import net.minecraft.core.Direction
import net.minecraft.core.SectionPos
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.level.block.entity.BlockEntity
import net.minecraftforge.common.capabilities.Capability
import net.minecraftforge.common.util.LazyOptional
import ru.dbotthepony.mc.otm.SERVER_IS_LIVE
import ru.dbotthepony.mc.otm.core.collect.WeakHashSet
import ru.dbotthepony.mc.otm.core.math.plus
import ru.dbotthepony.mc.otm.core.math.unaryMinus
import ru.dbotthepony.mc.otm.core.orNull
import ru.dbotthepony.mc.otm.onceServer
import java.lang.ref.WeakReference
import java.util.EnumMap
import java.util.function.Predicate

private inline val Direction.flag: Int get() = 1 shl (ordinal + 1)

/**
 * This class allows block entities to track [capability] per block side.
 *
 * Allowed sides are governed by [test] method, which is open for override.
 *
 * **IMPORTANT:** Once rules behind [test] are changed, you MUST call [validate] or [update] right away, or expect hard to debug bugs.
 * If rule changes can't be tracked in practical way, then flip on [alwaysValidate], although it will come with performance
 * penalty (this will cause [iterator] to invoke [validate] on each its invocation).
 *
 * Subclasses can also override [subscribe] and [unsubscribe] to more finely track changes in neighbours.
 *
 * **THIS CLASS IS COMPLETELY THREAD UNSAFE**, it is assumed you only ever invoke it on server thread,
 * or, at very least, thread where [ServerLevel] resides.
 */
open class BESubscribeList<T>(
	val block: BlockEntity,
	val capability: Capability<T>,
	val alwaysValidate: Boolean = false
) : Predicate<Direction>, Iterable<T> {
	private val knownCaps = WeakHashSet<LazyOptional<T>>()
	private val trackedCaps = EnumMap<Direction, WeakReference<LazyOptional<T>>>(Direction::class.java)
	private var updateQueued = false
	private var firstKey: Direction? = null
	private var lastDirectionSet = 0
	private var isWaitingOnChunk = false

	/**
	 * First valid capability, this is faster than invoking [iterator] and getting first value from it.
	 */
	val first: T? get() {
		return firstLO.orNull()
	}

	/**
	 * First valid capability as [LazyOptional], this is faster than invoking [iterator] and getting first value from it.
	 *
	 * If required, calls [update]
	 */
	val firstLO: LazyOptional<T> get() {
		if (updateQueued) {
			update()
		} else if (alwaysValidate) {
			validate()
		}

		for (i in 0 .. 2) {
			if (firstKey == null) {
				return LazyOptional.empty()
			}

			val get = trackedCaps[firstKey]?.get()

			if (get == null || !get.isPresent) {
				update()
				continue
			}

			return get
		}

		return LazyOptional.empty()
	}

	var valid = true
		set(value) {
			if (value != field) {
				field = value

				if (value) {
					update()
				}
			}
		}

	fun invalidate() {
		valid = false
	}

	fun revive() {
		valid = true
	}

	// override
	protected open fun subscribe(direction: Direction, capability: LazyOptional<T>) {}
	// override
	protected open fun unsubscribe(direction: Direction, capability: LazyOptional<T>?) {}

	override fun test(t: Direction): Boolean {
		return true
	}

	final override fun iterator(): Iterator<T> {
		if (updateQueued) {
			update()
		} else if (alwaysValidate) {
			validate()
		}

		return object : Iterator<T> {
			val iterator = trackedCaps.iterator()
			var value: T? = null

			private fun find() {
				while (iterator.hasNext() && value == null) {
					val (k, v) = iterator.next()
					value = v.get()?.orNull()

					if (value == null) {
						iterator.remove()
						unsubscribe(k, null)
					} else if (!test(k)) {
						iterator.remove()
						unsubscribe(k, null)
						updateQueued = true
					}
				}
			}

			override fun hasNext(): Boolean {
				find()
				return value != null
			}

			override fun next(): T {
				find()
				val value = value
				this.value = null
				return value ?: throw NoSuchElementException()
			}
		}
	}

	fun update(direction: Direction) {
		return update(direction::equals)
	}

	private fun calculateDirectionSet(): Int {
		var set = 0

		for (value in VALUES)
			if (test(value))
				set = set or value.flag

		return set
	}

	/**
	 * Checks for one of these conditions to be true:
	 * * If [test] was determined to be changed during last time [iterator] was invoked AND tracked capability was filtered out due to [test] returning false
	 * * Performs a fast bitflag intersection test, by calling [test] on all directions and seeking difference between built set and previous set checked during last [update]
	 * * Tracked capabilities are checked for validity (if references are still valid)
	 *
	 * Once *any* of conditions above end up true, [update] is called and nothing else down the list is checked further.
	 */
	fun validate() {
		check(valid) { "Subscription list is marked as invalid" }

		if (updateQueued) {
			update()
			return
		}

		val new = calculateDirectionSet()

		if (new != lastDirectionSet) {
			val intersect = new and lastDirectionSet
			val removed = lastDirectionSet and (intersect.inv())
			val added = new and (intersect.inv())

			for (dir in VALUES) {
				if (dir.flag and removed != 0) {
					val value1 = trackedCaps.remove(dir)
					val value = value1?.get()

					if (value1 != null) {
						unsubscribe(dir, value)
					}
				}
			}

			if (added != 0) {
				update { it.flag and added != 0 }
			}

			lastDirectionSet = new
			return
		}

		val iterator = trackedCaps.iterator()
		var any = 0

		for ((k, v) in iterator) {
			if (v.get() == null) {
				unsubscribe(k, null)
				any = any or k.flag
			}
		}

		if (any != 0) {
			update { it.flag and any != 0 }
		}
	}

	/**
	 * Updates subscription list, searching for new capabilities and "removing" outdated ones.
	 *
	 * [predicate] narrows sides-to-check list, not replaces the set generated by this class' [test]
	 */
	@JvmOverloads
	fun update(predicate: Predicate<Direction>? = null) {
		if (!valid || block.isRemoved)
			return

		val level = block.level as? ServerLevel ?: return // don't run on client
		val sorse = level.chunkSource
		var waitChunkLoad = false

		var stream = Direction.stream()

		if (predicate != null) {
			stream = stream.filter(predicate)
		} else {
			updateQueued = false
			lastDirectionSet = calculateDirectionSet()
		}

		stream = stream.filter(this)
		firstKey = null

		for (dir in stream) {
			val pos = block.blockPos + dir
			val getChunk = sorse.getChunkNow(SectionPos.blockToSectionCoord(pos.x), SectionPos.blockToSectionCoord(pos.z))

			if (getChunk == null) {
				waitChunkLoad = true
			} else {
				val cap = getChunk.getBlockEntity(pos)?.getCapability(capability, -dir)
				val knownCap = trackedCaps[dir]?.get()

				if (cap != null && cap.isPresent) {
					if (knownCap !== cap) {
						if (knownCaps.add(cap)) {
							val ref = WeakReference(this)

							cap.addListener {
								val self = ref.get()

								if (self != null) {
									val current = self.trackedCaps[dir]

									if (current?.get() === it) {
										self.update()
									}
								}
							}
						}

						trackedCaps[dir] = WeakReference(cap)
						subscribe(dir, cap)
					}

					if (firstKey == null) {
						firstKey = dir
					}
				} else {
					trackedCaps.remove(dir)
					unsubscribe(dir, knownCap)
				}
			}
		}

		if (waitChunkLoad && SERVER_IS_LIVE) {
			onceServer { if (isWaitingOnChunk) update() }
			isWaitingOnChunk = true
		} else if (predicate == null) {
			isWaitingOnChunk = false
		}
	}

	companion object {
		private val VALUES = Direction.values()
	}
}
